/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.activiti.timer;

import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.activiti.bpmn.model.BpmnModel;
import org.activiti.bpmn.model.Event;
import org.activiti.bpmn.model.EventDefinition;
import org.activiti.bpmn.model.FlowElement;
import org.activiti.bpmn.model.StartEvent;
import org.activiti.bpmn.model.TimerEventDefinition;
import org.activiti.engine.ActivitiIllegalArgumentException;
import org.activiti.engine.ProcessEngine;
import org.activiti.engine.delegate.event.ActivitiEntityEvent;
import org.activiti.engine.delegate.event.ActivitiEvent;
import org.activiti.engine.delegate.event.ActivitiEventListener;
import org.activiti.engine.impl.persistence.entity.AbstractJobEntity;
import org.activiti.engine.runtime.Job;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

public abstract class AbstractTimerRetryListener<T extends AbstractJobEntity> implements ActivitiEventListener {
    
    private final Logger logger = LoggerFactory.getLogger(this.getClass());
    private final Pattern isoRepeatPattern = Pattern.compile("^R(-?[0-9]*)/");
    
    @Autowired
    private ObjectMapper om;
    
    @Autowired
    protected ProcessEngine services;
    
    @Value("${inteligr8.timerExtension.alwaysRecordAttempts:false}")
    private boolean alwaysRecordAttempts;
    
    @Override
    public boolean isFailOnException() {
        return true;
    }
    
    protected boolean isEnabled(ActivitiEvent aevent, TimerEventDefinition timerDef, int repeats) {
        if (repeats < 0 && !this.alwaysRecordAttempts) {
            this.logger.debug("[pi:{}] The timer is not repeat limited, so not recording the number of tries: {}.{}",
                    aevent.getProcessInstanceId(), aevent.getProcessDefinitionId(), timerDef.getId());
            return false;
        }
        
        return true;
    }
    
    @Override
    public void onEvent(ActivitiEvent event) {
        this.logger.trace("onEvent({}, {}, {}, {})", event.getType(), event.getExecutionId(), event.getProcessInstanceId(), event.getClass());
        
        try {
            T job = this.getJobEntity(event);
            String activityId = this.getJobActivityId(job);
            this.logger.trace("Handling event for entity: {}.{}", event.getProcessDefinitionId(), activityId);
            
            Event modelEvent = this.getModelEvent(event.getProcessDefinitionId(), activityId);
            if (modelEvent == null || modelEvent instanceof StartEvent) {
                this.logger.trace("Start event detected; no special handling needed: {}", activityId);
                return;
            }

            this.onActivityEvent(event, job, modelEvent);
        } catch (IllegalArgumentException iae) {
            this.logger.warn(iae.getMessage());
        } catch (IllegalStateException ise) {
            this.logger.warn(ise.getMessage());
        } catch (JsonProcessingException jpe) {
            this.logger.warn(jpe.getMessage());
        }
    }
    
    public void onActivityEvent(ActivitiEvent event, T job, Event modelEvent) {
        // we are listening to timer events, so the event must have a timer definition
        TimerEventDefinition timerDef = this.getTimerDefinition(modelEvent);
        if (timerDef == null)
            throw new IllegalStateException("The event must have a timer definition, but none were found");
        this.onTimerEvent(event, job, modelEvent, timerDef);
    }
    
    public void onTimerEvent(ActivitiEvent event, T job, Event modelEvent, TimerEventDefinition timerDef) {
        int maxRepeats = this.determineRepeats(timerDef);
        if (maxRepeats < 0) {
            this.logger.debug("The '{}' timer is configured to repeat indefinitely", timerDef.getId());
        } else {
            this.logger.debug("The '{}' timer is configured to repeat {} times for {} total attempts", timerDef.getId(), maxRepeats, maxRepeats+1);
        }

        if (!this.isEnabled(event, timerDef, maxRepeats))
            return;
        
        this.onTimerEvent(event, job, modelEvent, timerDef, maxRepeats);
    }
    
    public void onTimerEvent(ActivitiEvent event, T job, Event modelEvent, TimerEventDefinition timerDef, int maxRepeats) {
        this.logger.trace("Attempts stored in variable: {}", this.getAttemptsVariableName(timerDef.getId()));
        
        int attempts = this.getAttempts(job, timerDef.getId());
        this.logger.debug("Timer executed {} of {} times in process instance: {}", attempts, maxRepeats+1, event.getProcessInstanceId());
        
        if (maxRepeats >= 0 && attempts > maxRepeats) {
            this.onTimerExhaustion(event, modelEvent, job);
        } else {
            this.onTimerAttempt(event, job, timerDef, attempts);
        }
    }
    
    public void onTimerExhaustion(ActivitiEvent event, Event modelEvent, T job) {
    }
    
    public abstract void onTimerAttempt(ActivitiEvent event, T job, TimerEventDefinition timerDef, int attempts);
    
    
    
    @SuppressWarnings("unchecked")
    protected T getJobEntity(ActivitiEvent event) {
        if (!(event instanceof ActivitiEntityEvent))
            throw new IllegalArgumentException("An entity event was expected; got: " + event.getClass());
        
        Object entity = ((ActivitiEntityEvent) event).getEntity();
        if (entity == null) {
            throw new IllegalStateException("An entity was expected; got: null");
        } else if (!(entity instanceof AbstractJobEntity)) {
            throw new IllegalStateException("An job entity was expected; got: " + entity.getClass());
        } else {
            return (T) entity;
        }
    }
    
    protected String getJobActivityId(Job job) throws JsonProcessingException {
        JsonNode configJson = this.om.readTree(job.getJobHandlerConfiguration());
        JsonNode activityIdJson = configJson.get("activityId");
        if (activityIdJson == null || activityIdJson.isNull())
            return null;
        return activityIdJson.textValue();
    }
    
    protected Event getModelEvent(String processDefinitionId, String elementId) {
        try {
            BpmnModel model = this.services.getRepositoryService().getBpmnModel(processDefinitionId);
            FlowElement element = model.getFlowElement(elementId);
            if (element == null) {
                throw new IllegalStateException("A model element was expected; got: null");
            } else if (!(element instanceof Event)) {
                throw new IllegalStateException("An event model element was expected; got: " + element.getClass());
            } else {
                return (Event) element;
            }
        } catch (ActivitiIllegalArgumentException aiae) {
            if (!aiae.getMessage().contains("process definition id is null"))
                throw aiae;
            
            this.logger.warn("The process is not yet available and the event must be a start event");
            return null;
        }
    }
    
    protected TimerEventDefinition getTimerDefinition(Event modelEvent) {
        Collection<EventDefinition> eventDefs = modelEvent.getEventDefinitions();
        if (eventDefs == null || eventDefs.isEmpty())
            return null;
        
        for (EventDefinition eventDef : eventDefs) {
            if (eventDef instanceof TimerEventDefinition)
                return (TimerEventDefinition) eventDef;
        }

        return null;
    }
    
    /**
     * @param timerDef
     * @return -1 for indefinition; 0 to never repeat; 1+ for number of repeats
     */
    protected int determineRepeats(TimerEventDefinition timerDef) {
        String cycle = timerDef.getTimeCycle();
        if (cycle == null)
            return -1;
        
        Matcher matcher = this.isoRepeatPattern.matcher(cycle);
        if (!matcher.find())
            return -1;
        
        String repeatValue = matcher.group(1);
        if (repeatValue.length() == 0) {
            return -1;
        } else {
            return Integer.parseInt(repeatValue);
        }
    }
    
    protected String getAttemptsVariableName(String timerId) {
        return timerId + ":attempts";
    }
    
    protected int getAttempts(Job job, String timerId) {
        Integer attempts = this.services.getRuntimeService().getVariable(
                job.getExecutionId(),
                this.getAttemptsVariableName(timerId),
                Integer.class);
        if (attempts == null)
            attempts = 0;
        return attempts.intValue();
    }
    
    protected void setAttempts(Job job, String timerId, int attempts) {
        this.services.getRuntimeService().setVariable(job.getExecutionId(),
                this.getAttemptsVariableName(timerId),
                attempts);
    }

}
