/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.activiti.timer;

import javax.annotation.PostConstruct;

import org.activiti.engine.ProcessEngine;
import org.activiti.engine.delegate.event.ActivitiEventType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

@Component
public class TimerExtension {
    
    @Autowired
    private ProcessEngine services;
    
    @Autowired
    private ScheduledTimerRetryListener scheduledListener;
    
    @Autowired
    private FiredTimerRetryListener firedListener;
    
    /**
     * By default, the attempt is incremented when the timer expires/fires.
     * That listener doesn't get called until the end of the regular execution.
     * If that execution ends the process instance, then setting a variable
     * results in wild SQL exceptions.  That is an Activiti bug.
     * 
     * The bug results in a rollback to the timer, eventually reaching the
     * deadletter state.  Setting this to true causes the increment to happen
     * when the timer is scheduled, not expired/fired.  This means the attempt
     * is recorded early, but we don't run into the bug.
     * 
     * Setting it to `false` causes the increment to happen after the attempt,
     * which may also be undesirable; and also must skirt the Activiti bug.
     */
    @Value("${inteligr8.timerExtension.recordAttemptOnScheduling:true}")
    private boolean recordAttemptOnScheduling;
    
    @PostConstruct
    public void registerListeners() {
        this.services.getRuntimeService().addEventListener(
                this.scheduledListener,
                ActivitiEventType.TIMER_SCHEDULED);
        if (!this.recordAttemptOnScheduling) {
            this.services.getRuntimeService().addEventListener(
                    this.firedListener,
                    ActivitiEventType.TIMER_FIRED);
        }
    }

}
